import dataclasses
from string import Template


class InvalidTemplate(AssertionError):
    pass


class InvalidTemplateKeys(InvalidTemplate):
    pass


@dataclasses.dataclass
class TemplateKeyInfo:
    template_keys: list[str]
    not_in_context: list[str]  # Keys are in template, but not in context
    not_in_template: list[str]  # Keys are in context, but not in template
    invalid_keys: list[str]


def get_template_key_info(*, content: str, context: dict) -> TemplateKeyInfo:
    template = Template(content)
    template_keys = set()
    invalid_keys = set()
    for mo in template.pattern.finditer(content):
        if name := (mo.group('named') or mo.group('braced')):
            template_keys.add(name)
        elif invalid := mo.group('invalid'):
            invalid_keys.add(invalid)

    context_keys = set(context.keys())
    return TemplateKeyInfo(
        template_keys=sorted(template_keys),
        not_in_template=sorted(context_keys - template_keys),
        not_in_context=sorted(template_keys - context_keys),
        invalid_keys=sorted(invalid_keys),
    )


def repr_keys(keys):
    """
    >>> repr_keys(['Foo', True, 1])
    '"Foo", "True", "1"'
    """
    return ', '.join(f'"{key}"' for key in keys)


def validate_template(*, content: str, context: dict) -> None:
    info: TemplateKeyInfo = get_template_key_info(content=content, context=context)

    errors = []
    if keys := info.not_in_context:
        errors.append(f'Template key(s) {repr_keys(keys)} not in context')

    if keys := info.not_in_template:
        errors.append(f'Context key(s) {repr_keys(keys)} not in template')

    if keys := info.invalid_keys:
        errors.append(f'Template key(s) {repr_keys(keys)} invalid')

    if errors:
        raise InvalidTemplate(' & '.join(errors))



import dataclasses
from pathlib import Path
from string import Template

from bx_py_utils.path import assert_is_dir, assert_is_file

from ha_services import __version__
from ha_services.cli_tools.richt_utils import human_error
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify
from ha_services.systemd.defaults import (
    get_demo_exec_start,
    get_template_path,
    get_user_group,
    get_user_name,
    get_work_directory,
)
from ha_services.systemd.template import InvalidTemplate, validate_template


@dataclasses.dataclass
class SystemdServiceTemplateContext:
    """
    Context values for the systemd service file content.
    Defaults are for the "DEMO publish-loop"
    """

    verbose_service_name: str = 'HaServices Demo'

    version: str = __version__
    user: str = dataclasses.field(default_factory=get_user_name)
    group: str = dataclasses.field(default_factory=get_user_group)
    work_dir: Path = dataclasses.field(default_factory=get_work_directory)

    exec_start: str = dataclasses.field(default_factory=get_demo_exec_start)

    # Optional values, will be automatically filled:
    syslog_identifier: str = None


@dataclasses.dataclass
class SystemdServiceInfo:
    """
    Information for systemd helper functions.
    Defaults are for the "DEMO publish-loop"
    """

    template_path: Path = dataclasses.field(default_factory=get_template_path)

    systemd_base_path: Path = Path('/etc/systemd/system/')

    # Set by post init:
    service_slug: str = None
    service_file_path: Path = None

    template_context: SystemdServiceTemplateContext = dataclasses.field(default_factory=SystemdServiceTemplateContext)

    def __post_init__(self):
        assert_is_dir(self.systemd_base_path)
        assert_is_dir(self.template_context.work_dir)
        assert_is_file(self.template_path)

        if not self.service_slug:
            self.service_slug = slugify(self.template_context.verbose_service_name, sep='_').lower()

        if not self.template_context.syslog_identifier:
            self.template_context.syslog_identifier = self.service_slug

        self.service_file_path = self.systemd_base_path / f'{self.service_slug}.service'

        try:
            validate_template(content=self.get_template_content(), context=self.get_template_context())
        except InvalidTemplate as err:
            human_error(
                f'Template {self.template_path} is not valid:\n{err}',
                title='[red]invalid Systemd template',
                exception=err,
            )

    def get_template_content(self) -> str:
        """
        Returns the content of template to generate the systemd service file
        """
        return self.template_path.read_text(encoding='UTF-8').strip()

    def get_template_context(self) -> dict:
        """
        Returns the context  for the systems service file template
        """
        context = dataclasses.asdict(self.template_context)
        return context

    def get_compiled_service(self) -> str:
        """
        Returns the completed systems service file content
        """
        template = self.get_template_content()
        template = Template(template)
        context = self.get_template_context()
        content = template.substitute(**context)
        return content



"""
    CLI for usage
"""
import logging
import sys
from pathlib import Path

import rich_click
import rich_click as click
from bx_py_utils.path import assert_is_file
from rich import print  # noqa
from rich.console import Console
from rich.traceback import install as rich_traceback_install
from rich_click import RichGroup

import ha_services
from ha_services import __version__, constants
from ha_services.example import DemoSettings, publish_forever
from ha_services.log_setup import basic_log_setup
from ha_services.mqtt4homeassistant.data_classes import MqttSettings
from ha_services.mqtt4homeassistant.mqtt import get_connected_client
from ha_services.systemd.api import ServiceControl
from ha_services.systemd.data_classes import SystemdServiceInfo
from ha_services.toml_settings.api import TomlSettings


logger = logging.getLogger(__name__)


PACKAGE_ROOT = Path(ha_services.__file__).parent.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')

OPTION_ARGS_DEFAULT_TRUE = dict(is_flag=True, show_default=True, default=True)
OPTION_ARGS_DEFAULT_FALSE = dict(is_flag=True, show_default=True, default=False)
ARGUMENT_EXISTING_DIR = dict(
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path)
)
ARGUMENT_NOT_EXISTING_DIR = dict(
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=False,
        writable=True,
        path_type=Path,
    )
)
ARGUMENT_EXISTING_FILE = dict(
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path)
)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


@click.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


cli.add_command(version)


######################################################################################################

SETTINGS_DIR_NAME = 'ha-services'
SETTINGS_FILE_NAME = 'ha-services-demo'


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def edit_settings(debug):
    """
    Edit the settings file. On first call: Create the default one.
    """
    basic_log_setup(debug=debug)
    TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    ).open_in_editor()


cli.add_command(edit_settings)


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def print_settings(debug):
    """
    Display (anonymized) MQTT server username and password
    """
    basic_log_setup(debug=debug)
    TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    ).print_settings()


cli.add_command(print_settings)


######################################################################################################
# Manage systemd service commands:


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def debug_systemd_config(debug):
    """
    Print Systemd service template + context + rendered file content.
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).debug_systemd_config()


cli.add_command(debug_systemd_config)


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def setup_systemd_service(debug):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).setup_and_restart_systemd_service()


cli.add_command(setup_systemd_service)


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def remove_systemd_service(debug):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).remove_systemd_service()


cli.add_command(remove_systemd_service)


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def systemd_status(debug):
    """
    Display status of systemd service. (May need sudo)
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).status()


cli.add_command(systemd_status)


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_TRUE)
def systemd_stop(debug):
    """
    Stops the systemd service. (May need sudo)
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    ServiceControl(info=systemd_settings).stop()


cli.add_command(systemd_stop)

######################################################################################################
# MQTT DEMO commands:


@click.command()
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_FALSE)
def test_mqtt_connection(debug):
    """
    Test connection to MQTT Server
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)

    settings: MqttSettings = user_settings.mqtt
    mqttc = get_connected_client(settings=settings, verbose=True)
    mqttc.loop_start()
    mqttc.loop_stop()
    mqttc.disconnect()
    print('\n[green]Test succeed[/green], bye ;)')


cli.add_command(test_mqtt_connection)


@click.command()
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--debug/--no-debug', **OPTION_ARGS_DEFAULT_FALSE)
def publish_loop(verbose, debug):
    """
    Publish data via MQTT for Home Assistant (endless loop)
    """
    basic_log_setup(debug=debug)
    toml_settings = TomlSettings(
        dir_name=SETTINGS_DIR_NAME,
        file_name=SETTINGS_FILE_NAME,
        settings_dataclass=DemoSettings(),
    )
    user_settings: DemoSettings = toml_settings.get_user_settings(debug=True)

    try:
        publish_forever(user_settings=user_settings, verbose=verbose)
    except KeyboardInterrupt:
        print('Bye, bye')


cli.add_command(publish_loop)


######################################################################################################


def main():
    print(f'[bold][green]ha-services[/green] DEMO cli v[cyan]{__version__}')

    console = Console()
    rich_traceback_install(
        width=console.size.width,  # full terminal width
        show_locals=True,
        suppress=[click, rich_click],
        max_frames=2,
    )

    # Execute Click CLI:
    cli.name = './cli.py'
    cli()



from __future__ import annotations

import sys

from bx_py_utils.test_utils.assertion import text_unified_diff
from pygments.lexer import Lexer
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.panel import Panel
from rich.pretty import Pretty
from rich.syntax import Syntax
from rich.text import Text


def print_code(
    code: str,
    *,
    title: str | None = None,
    lexer: Lexer | str,
    theme: str = 'ansi_dark',
    background_color: str = '#090909',
    line_numbers: bool = True,
    console: Console | None = None,
) -> None:
    """
    Print source code via Pygments
    """
    console = console or Console()

    console.print()
    console.rule(title)
    console.print(
        Syntax(
            code,
            lexer,
            theme=theme,
            background_color=background_color,
            line_numbers=line_numbers,
        )
    )
    console.rule()
    console.print()


def print_unified_diff(
    txt1: str,
    txt2: str,
    *,
    fromfile: str = 'got',
    tofile: str = 'expected',
    title: str = '[bright][green]unified diff[/green]:',
    console: Console | None = None,
) -> None:
    """
    Print a diff highlight with Pygments
    """
    diff = text_unified_diff(
        txt1=txt1,
        txt2=txt2,
        fromfile=fromfile,
        tofile=tofile,
    )
    print_code(code=diff, lexer='diff', title=title, console=console)


class PanelPrinter:
    def __init__(
        self,
        *,
        HighlighterClass=ReprHighlighter,
        border_style='bright_yellow',
        padding=(2, 5),
        console=None,
    ):
        if HighlighterClass is not None:
            self.highlighter = HighlighterClass()
        else:
            self.highlighter = None

        self.HighlighterClass = HighlighterClass
        self.border_style = border_style
        self.padding = padding
        self.console = console or Console()
        self.console.print()

    def print_panel(self, *, content, title, border_style=None):
        self.console.print(
            Panel(
                self.highlight(content),
                title=title,
                border_style=border_style or self.border_style,
                expand=False,
                padding=self.padding,
            )
        )
        self.console.print()

    def highlight(self, content):
        if not isinstance(content, (str, Text)):
            content = Pretty(content)
        elif self.highlighter:
            content = self.highlighter(content)
        return content


def human_error(
    message: str,
    *,
    title='[red]Error',
    exit_code=None,
    exception: BaseException | None = None,
    exception_extra_lines: int = 2,
    exception_max_frames: int = 20,
    HighlighterClass=ReprHighlighter,
    border_style='bright_red',
    padding=(2, 5),
    console=None,
):
    """
    Print a message in a Panel.
    Optional:
        - print traceback first
        - Exit with: sys.exit(exit_code)
    """
    if console is None:
        use_stderr = exception is not None
        console = Console(stderr=use_stderr)

    console.print('\n')

    if exception is not None:
        assert isinstance(exception, BaseException), f'Not a exception: {exception!r}'
        console.print_exception(
            width=console.size.width,  # full terminal width
            extra_lines=exception_extra_lines,
            show_locals=True,
            max_frames=exception_max_frames,
        )

    console.print()

    pp = PanelPrinter(HighlighterClass=HighlighterClass, border_style=border_style, padding=padding, console=console)
    pp.print_panel(content=message, title=title)

    if exit_code is not None:
        sys.exit(exit_code)



import tempfile
from pathlib import Path
from unittest import TestCase

from ha_services.cli_tools.path_utils import backup, expand_user
from ha_services.cli_tools.test_utils.environment_fixtures import AsSudoCallOverrideEnviron


class PathUtilsTestCase(TestCase):
    def test_expand_user(self):
        real_home_path = Path.home()
        real_example = Path('~/example/').expanduser()

        self.assertEqual(expand_user(Path('~')), real_home_path)
        self.assertEqual(expand_user(Path('~/example/')), real_example)

        with AsSudoCallOverrideEnviron():
            self.assertEqual(Path('~').expanduser(), Path('/root'))
            self.assertEqual(expand_user(Path('~')), real_home_path)

            self.assertEqual(Path('~/example/').expanduser(), Path('/root/example'))
            self.assertEqual(expand_user(Path('~/example/')), real_example)

    def test_backup(self):
        with tempfile.TemporaryDirectory(prefix='test_') as temp_dir:
            temp_path = Path(temp_dir)

            test_path = temp_path / 'foobar.ext'
            test_path.write_text('one')

            bak1_path = backup(test_path)
            self.assertEqual(bak1_path, temp_path / 'foobar.ext.bak')
            self.assertEqual(bak1_path.read_text(), 'one')

            test_path.write_text('two')

            bak2_path = backup(test_path)
            self.assertEqual(bak2_path, temp_path / 'foobar.ext.bak2')
            self.assertEqual(bak2_path.read_text(), 'two')

            test_path.write_text('three')

            bak3_path = backup(test_path)
            self.assertEqual(bak3_path, temp_path / 'foobar.ext.bak3')
            self.assertEqual(bak3_path.read_text(), 'three')

            with self.assertRaises(RuntimeError) as cm:
                backup(test_path, max_try=3)
            self.assertEqual(cm.exception.args, ('No backup made: Maximum attempts to find a file name failed.',))



from click._compat import strip_ansi as strip_ansi_codes
from rich.console import Console

from ha_services.cli_tools.richt_utils import PanelPrinter


def assert_in(content: str, parts: tuple[str, ...], strip_ansi=True) -> None:
    """
    Check if all parts exist in content
    """
    if strip_ansi:
        content = strip_ansi_codes(content)

    missing = [part for part in parts if part not in content]
    if missing:
        console = Console()
        console.rule(title='assert_in(): Content start', characters='∨')
        print(content)
        console.rule(title='assert_in(): Content end', characters='∧')

        pp = PanelPrinter()
        pp.print_panel(
            title='assert_in(): [red]Missing parts:',
            content='\n\n'.join(missing),
        )
        raise AssertionError(f'assert_in(): {len(missing)} parts not found in content, see output above')


def assert_startswith(text, prefix):
    if not text.startswith(prefix):
        raise AssertionError(f'{text!r} does not starts with: {prefix!r}')



from pathlib import Path
from unittest import TestCase

from tomlkit import TOMLDocument

from ha_services.example import DemoSettings
from ha_services.systemd.data_classes import SystemdServiceInfo
from ha_services.toml_settings.deserialize import toml2dataclass
from ha_services.toml_settings.serialize import dataclass2toml


class TomlSettingsTestCase(TestCase):
    def test_demo_settings(self):
        document = dataclass2toml(instance=DemoSettings())
        self.assertIsInstance(document, TOMLDocument)
        data = document.unwrap()

        self.assertEqual(data['app']['device_name'], 'ha-services-demo')
        self.assertEqual(data['mqtt']['host'], 'mqtt.eclipseprojects.io')
        self.assertEqual(data['systemd']['service_slug'], 'haservices_demo')
        self.assertEqual(data['systemd']['systemd_base_path'], '/etc/systemd/system')

        instance = DemoSettings(systemd=SystemdServiceInfo(service_slug='foo bar'))
        instance.systemd.systemd_base_path = Path('/foo/bar')

        toml2dataclass(document=document, instance=instance)

        self.assertEqual(instance.systemd.systemd_base_path, Path('/etc/systemd/system'))



import dataclasses
import inspect
import logging
from pathlib import Path
from unittest import TestCase

import tomlkit

from ha_services.toml_settings.deserialize import toml2dataclass
from ha_services.toml_settings.tests.fixtures import ComplexExample, PathExample, SimpleExample


class DeserializeTestCase(TestCase):
    def test_toml2dataclass_simple(self):
        instance = SimpleExample()
        data = dataclasses.asdict(instance)
        self.assertEqual(data, {'one': 'foo', 'two': 'bar', 'three': '', 'number': 123})

        document = tomlkit.loads(
            inspect.cleandoc(
                '''
                # Own Comment,
                # should be keep
                two = "New Value"
                three = 666 # Wrong type!
                number = 123 # Contains the same value!

                [other]
                foo = "bar"
                '''
            ),
        )
        with self.assertLogs(logger=None, level=logging.DEBUG) as logs:
            toml2dataclass(document=document, instance=instance)
        print('\n'.join(logs.output))

        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'one': 'foo',
                'two': 'New Value',  # <<< changed
                'three': '',
                'number': 123,
            },
        )
        toml = tomlkit.dumps(document)
        self.assertEqual(
            toml,
            inspect.cleandoc(
                '''
                # Own Comment,
                # should be keep
                two = "New Value"
                three = "" # Wrong type!
                number = 123 # Contains the same value!
                one = "foo"

                [other]
                foo = "bar"
                '''
            ),
        )

        self.assertEqual(
            logs.output,
            [
                "INFO:ha_services.toml_settings.deserialize:Missing 'one' in toml config",
                #
                "INFO:ha_services.toml_settings.deserialize:Take over 'two' from user toml setting",
                #
                'ERROR:ha_services.toml_settings.deserialize:Toml value three=666 is type '
                "'int' but must be type 'str' -> ignored and use default value!",
                #
                'DEBUG:ha_services.toml_settings.deserialize:Default value 123 also used in toml file, ok.',
            ],
        )

    def test_toml2dataclass_path(self):
        instance = PathExample()
        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'path': Path('/foo/bar'),
            },
        )

        document = tomlkit.loads(
            inspect.cleandoc(
                '''
                path = "/to/some/other/place/"
                '''
            ),
        )

        toml2dataclass(document=document, instance=instance)

        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'path': Path('/to/some/other/place'),
            },
        )

    def test_toml2dataclass_inheritance(self):
        instance = ComplexExample()
        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'foo': 'bar',
                'sub_class_one': {'number': 123},
                'sub_class_two': {'something': 0.5},
                'sub_class_three': {'one_value': True},
            },
        )

        document = tomlkit.loads(
            inspect.cleandoc(
                '''
                [sub_class_one]
                number = "Not a Number!"

                [sub_class_two]
                something = 123.456
                '''
            ),
        )
        with self.assertLogs(logger=None, level=logging.DEBUG) as logs:
            toml2dataclass(document=document, instance=instance)
        print('\n'.join(logs.output))

        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'foo': 'bar',  # <<< not change, because not in toml file
                'sub_class_one': {
                    'number': 123,  # <<< not change, because wrong type
                },
                'sub_class_two': {
                    'something': 123.456,  # <<< updated
                },
                'sub_class_three': {'one_value': True},  # <<< added
            },
        )
        toml = tomlkit.dumps(document).rstrip()
        self.assertEqual(
            toml,
            inspect.cleandoc(
                '''
                foo = "bar"
                [sub_class_one]
                number = 123

                [sub_class_two]
                something = 123.456

                [sub_class_three]
                # SubClass3(one_value: bool = True)
                one_value = true
                '''
            ),
        )

        self.assertEqual(
            logs.output,
            [
                "INFO:ha_services.toml_settings.deserialize:Missing 'foo' in toml config",
                #
                "ERROR:ha_services.toml_settings.deserialize:Toml value number='Not a "
                "Number!' is type 'str' but must be type 'int' -> ignored and use default "
                'value!',
                #
                "INFO:ha_services.toml_settings.deserialize:Take over 'something' from user toml setting",
                #
                'INFO:ha_services.toml_settings.deserialize:Missing complete sub dataclass '
                "'sub_class_three' in toml config",
            ],
        )



import getpass
import os
import shlex
import sys

from bx_py_utils.environ import OverrideEnviron


class AsSudoCallOverrideEnviron(OverrideEnviron):
    """
    Manipulate the environment variables so it looks like a "sudo" call.
    Compare with: "sudo env | sort" ;)
    """

    def __init__(self, **overrides):
        overrides = {
            'SUDO_COMMAND': shlex.join(sys.argv),
            'SUDO_UID': str(os.getuid()),
            'SUDO_GID': str(os.getgid()),
            'SUDO_USER': getpass.getuser(),
            'LOGNAME': 'root',
            'HOME': '/root',
            **overrides,
        }
        super().__init__(**overrides)



import logging


def basic_log_setup(*, debug: bool, format='%(asctime)s %(levelname)-8s %(message)s'):
    if debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(level=level, format=format)



import dataclasses
import inspect
import textwrap
from collections.abc import Iterable

from bx_py_utils.anonymize import anonymize
from rich import print  # noqa
from rich.console import Console
from rich.style import Style

from ha_services.toml_settings.data_class_utils import iter_dataclass


def print_dataclasses(
    *, instance, anonymize_keys: Iterable, indent=0, style=Style(bgcolor='#090909'), console=None
) -> set[str]:
    console = console or Console()

    console.print(f'{" "*indent}[magenta]{instance.__class__.__name__}[/magenta]:', style=style)

    indent += 2

    if doc_string := inspect.getdoc(instance):
        console.print(textwrap.indent(doc_string, prefix=' ' * indent), style=style)

    indent += 2

    anonymized = set()

    for field_name, field_value in iter_dataclass(instance):
        if dataclasses.is_dataclass(field_value):
            console.print()
            anonymized |= print_dataclasses(
                instance=field_value, anonymize_keys=anonymize_keys, indent=indent + 4, style=style, console=console
            )
        else:
            if field_name in anonymize_keys:
                field_value = anonymize(field_value)
                anonymized.add(field_name)
            console.print(f'{" "*indent}* [cyan]{field_name}[/cyan] = {field_value!r}', style=style)
    return anonymized



import logging
from subprocess import CalledProcessError

from manageprojects.utilities.subprocess_utils import verbose_check_call
from rich import print  # noqa
from rich.console import Console
from rich.highlighter import ReprHighlighter

from ha_services.cli_tools.richt_utils import PanelPrinter, human_error, print_code, print_unified_diff
from ha_services.systemd.data_classes import SystemdServiceInfo


logger = logging.getLogger(__name__)


class SystemdServiceError(RuntimeError):
    pass


class ServiceControl:
    """
    Manage Systemd service
    """

    def __init__(self, info: SystemdServiceInfo):
        self.info = info
        self.service_name = info.service_file_path.name

    ##################################################################################################
    # Helper

    def sudo_hint_exception_exit(self, err, exit_code=1):
        human_error(
            f'{err}\n\nHint: Maybe **sudo** is needed for this command!\nTry again with sudo.',
            title='[red]Permission error',
            exception=err,
            exit_code=exit_code,
        )

    def write_service_file(self):
        print(f'Write "{self.info.service_file_path}"...')
        content = self.info.get_compiled_service()
        try:
            self.info.service_file_path.write_text(content, encoding='UTF-8')
        except PermissionError as err:
            self.sudo_hint_exception_exit(err)

        self.call_systemctl('daemon-reload', with_service_name=False)

    def remove_service_file(self):
        print(f'Remove "{self.info.service_file_path}"...')
        try:
            self.info.service_file_path.unlink(missing_ok=True)
        except PermissionError as err:
            self.sudo_hint_exception_exit(err)

        self.call_systemctl('daemon-reload', with_service_name=False)

    def service_file_is_up2date(self, print_warning: bool = True, print_diff: bool = False, console=None):
        logger.debug('Read %s...', self.info.service_file_path)
        try:
            got_content = self.info.service_file_path.read_text(encoding='UTF-8')
        except PermissionError as err:
            human_error(
                title=f'[yellow]WARNING - {self.info.service_file_path}',
                message=f'Can not read the system service file, to check the content:\n{err}',
                border_style='yellow',
            )
        else:
            expect_content = self.info.get_compiled_service()
            if got_content.strip() == expect_content.strip():
                logger.info('%s is up2date, ok.', self.info.service_file_path)
                return True
            else:
                logger.warning('%s is not up2date!', self.info.service_file_path)
                if print_diff:
                    print_unified_diff(
                        txt1=expect_content,
                        txt2=got_content,
                        fromfile=str(self.info.service_file_path),
                        tofile='Fresh compiled with current settings',
                        title=f'[bright][green]unified diff[/green]: {self.info.service_file_path}',
                    )
                elif print_warning:
                    human_error(
                        title=f'[yellow]WARNING - {self.info.service_file_path}',
                        message='The systemd service file is not up2date with the current settings!',
                        border_style='yellow',
                    )
                return False

    ##################################################################################################
    # systemctl

    def call_systemctl(self, command, with_service_name=True, exit_on_error=True):
        args = ['systemctl', command]
        if with_service_name:
            args.append(self.service_name)
            if not self.info.service_file_path.is_file():
                human_error(
                    f'Systemd service file not found here: {self.info.service_file_path}'
                    '\n\nHint: Setup systemd service first!',
                    title='[red]Missing systems service file',
                    exit_code=1,
                )
            else:
                self.service_file_is_up2date(print_warning=True)

        try:
            verbose_check_call(*args)
        except CalledProcessError as err:
            if exit_on_error:
                self.sudo_hint_exception_exit(err)
            raise

    def enable(self):
        self.call_systemctl('enable')

    def restart(self):
        self.call_systemctl('restart')

    def stop(self):
        self.call_systemctl('stop')

    def status(self):
        try:
            self.call_systemctl('status', exit_on_error=False)
        except CalledProcessError as err:
            # Exit code is not 0 if systemd service is not running!
            human_error(
                message=(
                    'Hints:\nSystemd service is just not running, cause of a error?!\nTry to setup the services.'
                    '\nCheck your template context.'
                ),
                title=f'[bold]{err}',
                border_style='bright_yellow',
                exception=None,  # Don't print traceback
                exit_code=err.returncode,
            )

    ##################################################################################################
    # High level commands:
    def setup_and_restart_systemd_service(self):
        """
        Write Systemd service file, enable it and (re-)start the service.
        """
        self.write_service_file()
        self.enable()
        self.restart()
        self.status()

    def remove_systemd_service(self):
        self.stop()
        self.remove_service_file()

    ##################################################################################################
    # Debug

    def debug_systemd_config(
        self,
        *,
        print_diff: bool = True,
        HighlighterClass=ReprHighlighter,
        padding=(1, 5),
        console=None,
    ):
        """
        Print Systemd service template + context + rendered file content.
        """
        console = console or Console()
        pp = PanelPrinter(
            HighlighterClass=HighlighterClass,
            border_style='white',
            padding=padding,
        )
        pp.print_panel(
            content=self.info.get_template_content(),
            title=f'[magenta]Template[/magenta]: {self.info.template_path}',
        )
        pp.print_panel(
            content=self.info.get_template_context(),
            title='[cyan]Context:',
        )

        print_code(
            code=self.info.get_compiled_service(),
            lexer='ini',
            title=f'[bright][green]Compiled[/green]: {self.info.service_file_path.name}',
            console=console,
        )

        if self.info.service_file_path.is_file():
            if not print_diff:
                # Just print a warning, if service file is not up2date:
                self.service_file_is_up2date(print_warning=True)
            else:
                self.service_file_is_up2date(print_diff=True, console=console)
        else:
            logger.debug('Systemd service not created -> skip up2date check.')



from unittest import TestCase
from unittest.mock import patch

from bx_py_utils.path import assert_is_file
from bx_py_utils.test_utils.redirect import RedirectOut
from manageprojects.test_utils.subprocess import SubprocessCallMock
from manageprojects.utilities import subprocess_utils

from ha_services import __version__
from ha_services.cli_tools.test_utils.assertion import assert_in
from ha_services.systemd.api import ServiceControl
from ha_services.systemd.tests.utilities import MockedSystemdServiceInfo


class MockedShutilWhich:
    def which(self, command, path=None):
        return f'/usr/bin/{command}'


class SystemdApiTestCase(TestCase):
    def test_print_systemd_file(self):
        with MockedSystemdServiceInfo(prefix='test_print_systemd_file_') as info, RedirectOut() as buffer:
            ServiceControl(info=info).debug_systemd_config()

        self.assertEqual(buffer.stderr, '')
        assert_in(
            content=buffer.stdout,
            parts=(
                '[Unit]',
                f'Description=HaServices Demo {__version__}',
                'ExecStart=/mocked/.venv/bin/python3 -m ha_services_app publish-loop',
                'SyslogIdentifier=haservices_demo',
            ),
        )

    def test_service_control(self):
        with MockedSystemdServiceInfo(prefix='test_') as info:
            service_control = ServiceControl(info=info)

            for func_name in ('enable', 'restart', 'stop', 'status', 'remove_systemd_service'):
                with self.subTest(func_name):
                    service_control_func = getattr(service_control, func_name)
                    with RedirectOut() as buffer, self.assertRaises(SystemExit):
                        service_control_func()
                    assert_in(
                        content=buffer.stdout,
                        parts=(
                            'Systemd service file not found',
                            'Hint: Setup systemd service first!',
                        ),
                    )

            with SubprocessCallMock() as mock, patch.object(
                subprocess_utils, 'shutil', MockedShutilWhich()
            ), RedirectOut() as buffer:
                service_control.setup_and_restart_systemd_service()

            assert_in(
                content=buffer.stdout,
                parts=(
                    f'Write "{info.service_file_path}"...',
                    'systemctl daemon-reload',
                    'systemctl enable haservices_demo.service',
                    'systemctl restart haservices_demo.service',
                    'systemctl status haservices_demo.service',
                ),
            )
            assert_is_file(info.service_file_path)

            self.assertEqual(
                mock.get_popenargs(),
                [
                    ['/usr/bin/systemctl', 'daemon-reload'],
                    ['/usr/bin/systemctl', 'enable', 'haservices_demo.service'],
                    ['/usr/bin/systemctl', 'restart', 'haservices_demo.service'],
                    ['/usr/bin/systemctl', 'status', 'haservices_demo.service'],
                ],
            )



import dataclasses
import logging
from collections.abc import Iterable
from pathlib import Path

import tomlkit
from rich import print  # noqa
from rich.console import Console
from tomlkit import TOMLDocument

from ha_services.cli_tools.path_utils import backup, expand_user
from ha_services.cli_tools.richt_utils import human_error
from ha_services.toml_settings.debug import print_dataclasses
from ha_services.toml_settings.deserialize import toml2dataclass
from ha_services.toml_settings.exceptions import UserSettingsNotFound
from ha_services.toml_settings.sensible_editor import open_editor_for
from ha_services.toml_settings.serialize import dataclass2toml


logger = logging.getLogger(__name__)


class TomlSettings:
    settings_directories = (  # Path candidates where setting files will be stored
        '~/.config/',
        '~',
    )

    # Print this error message if settings file not exists:
    not_exists_hint = '[yellow]No settings created yet[/yellow]: [green](Hint: call "edit-settings" first!)'

    def __init__(
        self,
        *,
        dir_name: str,
        file_name: str,
        settings_dataclass: dataclasses,
        not_exist_exit_code=-1,
    ):
        self.file_path = self.get_settings_file_path(dir_name, file_name)

        assert dataclasses.is_dataclass(settings_dataclass), f'No dataclass: {settings_dataclass}'
        self.settings_dataclass = settings_dataclass
        self.not_exist_exit_code = not_exist_exit_code

    def get_settings_path(self) -> Path:
        candidates = []
        for candidate in self.settings_directories:
            if path := expand_user(Path(candidate)):  # Expand with user home, even if called via sudo!
                candidates.append(path)
                if path.is_dir():
                    return path
        raise RuntimeError(f'All settings directories does not exists! Tried: {candidates}')

    def get_settings_file_path(self, dir_name: str, file_name: str) -> Path:
        assert dir_name == Path(dir_name).stem, f'Invalid {dir_name=!r}'
        assert file_name == Path(file_name).stem, f'Invalid {file_name=!r}'
        settings_path = self.get_settings_path() / dir_name
        if not settings_path.exists():
            settings_path.mkdir(parents=False, exist_ok=False)
        return settings_path / f'{file_name}.toml'

    def open_in_editor(self) -> None:
        if not self.file_path.is_file():
            logger.info('Settings file "%s" not exist -> create default', self.file_path)
            document: TOMLDocument = dataclass2toml(instance=self.settings_dataclass)
            doc_str = tomlkit.dumps(document, sort_keys=False)
            self.file_path.write_text(doc_str, encoding='UTF-8')

        open_editor_for(self.file_path)

    def get_user_settings(self, *, debug: bool = False) -> dataclasses:
        if debug:
            print(f'Use user settings file: {self.file_path}', end='...')
        if not self.file_path.is_file():
            human_error(message=self.not_exists_hint, exit_code=self.not_exist_exit_code)
            raise UserSettingsNotFound(self.file_path)

        doc_str = self.file_path.read_text(encoding='UTF-8')
        user_settings_doc: TOMLDocument = tomlkit.loads(doc_str)

        document_changed = toml2dataclass(document=user_settings_doc, instance=self.settings_dataclass)
        logger.debug(f'{document_changed=}')
        if document_changed:
            logger.info('User toml file needs update!')
            doc_str = tomlkit.dumps(user_settings_doc, sort_keys=False)
            backup(self.file_path)
            self.file_path.write_text(doc_str, encoding='UTF-8')

        if debug:
            print('[green]read, ok.')
        return self.settings_dataclass

    def print_settings(self, *, anonymize_keys: Iterable = ('password', 'email')) -> None:
        """
        Print (anonymized) settings
        """
        user_settings = self.get_user_settings(debug=True)

        print()
        console = Console()
        console.rule(f'User settings: {self.file_path}')
        anonymized = print_dataclasses(instance=user_settings, anonymize_keys=anonymize_keys)
        if anonymized:
            info = f'(anonymize keys: {", ".join(sorted(anonymized))})'
        else:
            info = None
        console.rule(info)



from unittest import TestCase

from ha_services.mqtt4homeassistant.converter import values2mqtt_payload
from ha_services.mqtt4homeassistant.data_classes import HaMqttPayload, HaValue, HaValues


class ConverterTestCase(TestCase):
    def test_basic(self):
        values = HaValues(
            device_name='egg boiler',
            values=[
                HaValue(
                    name='Power',
                    value=450,
                    device_class='energy',
                    state_class='measurement',
                    unit='W',
                )
            ],
            prefix='homeassistant',
            component='sensor',
        )
        self.assertEqual(
            values2mqtt_payload(values=values, name_prefix='Test'),
            HaMqttPayload(
                configs=[
                    {
                        'data': {
                            'device': {
                                'identifiers': ['test_eggboiler_power'],
                                'name': 'egg boiler',
                            },
                            'device_class': 'energy',
                            'name': 'Power',
                            'state_class': 'measurement',
                            'state_topic': 'homeassistant/sensor/test_eggboiler/state',
                            'unique_id': 'test_eggboiler_power',
                            'unit_of_measurement': 'W',
                            'value_template': '{{ value_json.test_eggboiler_power }}',
                        },
                        'topic': 'homeassistant/sensor/test_eggboiler_power/config',
                    }
                ],
                state={
                    'data': {
                        'test_eggboiler_power': 450,
                    },
                    'topic': 'homeassistant/sensor/test_eggboiler/state',
                },
            ),
        )

    def test_check_prefix(self):
        with self.assertRaises(AssertionError) as err:
            values2mqtt_payload(values=None, name_prefix='1foobar'),
        self.assertIn('Invalid: name_prefix', str(err.exception))



import dataclasses
import logging
from pathlib import Path

import tomlkit
from tomlkit import TOMLDocument

from ha_services.toml_settings.data_class_utils import iter_dataclass
from ha_services.toml_settings.serialize import add_dataclass


logger = logging.getLogger(__name__)


def toml2dataclass(*, document: TOMLDocument, instance, _changed=False) -> bool:
    """
    Sync toml and instance in place:

     - Transfer all valid values from toml document into the data class instance
     - Add missing or invalid values to the toml document
    """

    for field_name, field_value in iter_dataclass(instance):
        if dataclasses.is_dataclass(field_value):
            try:
                sub_doc = document[field_name]
            except KeyError:
                logger.info('Missing complete sub dataclass %r in toml config', field_name)
                document.add(tomlkit.nl())  # Add new line
                add_dataclass(document=document, name=field_name, instance=field_value)
                _changed = True
            else:
                changed = toml2dataclass(document=sub_doc, instance=field_value, _changed=_changed)
                if changed:
                    _changed = True
            continue

        try:
            doc_value = document[field_name]
        except KeyError:
            logger.info('Missing %r in toml config', field_name)

            # Add missing item, so that the toml file can be updated on disk:
            document[field_name] = field_value
            _changed = True
            continue

        if field_value == doc_value:
            logger.debug('Default value %r also used in toml file, ok.', field_value)
        elif isinstance(field_value, Path):
            logger.debug('Convert %r value %r to Path instance', field_name, field_value)
            value = doc_value.unwrap()
            if not isinstance(value, str):
                logger.error(
                    (
                        'Toml value %s=%r is type %r but must be type str (Will be convert to Path instance)'
                        ' -> ignored and use default value!'
                    ),
                    field_name,
                    value,
                    type(value).__name__,
                )
                document[field_name] = field_value  # Add default one
                _changed = True
            else:
                setattr(instance, field_name, Path(value))
        elif not isinstance(field_value, type(doc_value.unwrap())):
            logger.error(
                'Toml value %s=%r is type %r but must be type %r -> ignored and use default value!',
                field_name,
                doc_value,
                type(doc_value.unwrap()).__name__,
                type(field_value).__name__,
            )
            document[field_name] = field_value  # Add default one
            _changed = True
        else:
            logger.info('Take over %r from user toml setting', field_name)
            setattr(instance, field_name, doc_value)

    return _changed



import json
import logging
import socket

import paho.mqtt.client as mqtt
from bx_py_utils.anonymize import anonymize
from rich import print
from rich.pretty import pprint

from ha_services import __version__
from ha_services.cli_tools.richt_utils import human_error
from ha_services.mqtt4homeassistant.data_classes import HaMqttPayload, MqttSettings


logger = logging.getLogger(__name__)


def get_client_id():
    hostname = socket.gethostname()
    client_id = f'ha_services v{__version__} on {hostname}'
    return client_id


class OnConnectCallback:
    def __init__(self, verbose: bool = True):
        self.verbose = verbose

    def __call__(self, client, userdata, flags, rc):
        if self.verbose:
            print(f'MQTT broker connect result code: {rc}', end=' ')

        if rc == 0:
            if self.verbose:
                print('[green]OK')
        else:
            print('\n[red]MQTT Connection not successful!')
            print('[yellow]Please check your credentials\n')
            raise RuntimeError(f'MQTT connection result code {rc} is not 0')

        if self.verbose:
            print(f'\t{userdata=}')
            print(f'\t{flags=}')


def get_connected_client(settings: MqttSettings, verbose: bool = True, timeout=10):
    client_id = get_client_id()

    if verbose:
        print(f'\nConnect [cyan]{settings.host}:{settings.port}[/cyan] as "[magenta]{client_id}[/magenta]"', end='...')

    socket.setdefaulttimeout(timeout)  # Sadly: Timeout will not used in getaddrinfo()!
    try:
        info = socket.getaddrinfo(settings.host, settings.port)
    except socket.gaierror as err:
        human_error(
            message=f'{err}\n(Hint: Check you host/port settings)',
            title=f'[red]Error get address info from: [cyan]{settings.host}:{settings.port}[/cyan]',
            exception=err,
            exit_code=1,
        )
    else:
        if not info:
            print('[red]Resolve error: No info!')
        elif verbose:
            print('Host/port test [green]OK')

    mqttc = mqtt.Client(client_id=client_id)
    mqttc.on_connect = OnConnectCallback(verbose=verbose)
    mqttc.enable_logger(logger=logger)

    if settings.user_name and settings.password:
        if verbose:
            print(
                f'login with user: {anonymize(settings.user_name)} password:{anonymize(settings.password)}',
                end='...',
            )
        mqttc.username_pw_set(settings.user_name, settings.password)

    mqttc.connect(settings.host, port=settings.port)

    if verbose:
        print('[green]OK')
    return mqttc


class HaMqttPublisher:
    def __init__(self, settings: MqttSettings, verbose: bool = True, config_count: int = 10):
        self.verbose = verbose
        self.mqttc = get_connected_client(settings=settings, verbose=verbose)
        self.mqttc.loop_start()

        self.config_count = config_count
        self.send_count = 0

    def publish(self, *, topic: str, payload: dict) -> None:
        if self.verbose:
            print('_' * 100)
            print(f'[yellow]Publish MQTT topic: [blue]{topic} [grey](Send count: {self.send_count})')
            pprint(payload)

        info: mqtt.MQTTMessageInfo = self.mqttc.publish(topic=topic, payload=json.dumps(payload))
        if self.verbose:
            print('publish result:', info)

    def publish2homeassistant(self, *, ha_mqtt_payload: HaMqttPayload) -> None:
        log_prefix = f'{self.send_count=} ({self.config_count=})'

        if self.send_count == 0 or self.send_count % self.config_count == 0:
            logger.debug(f'{log_prefix} send {len(ha_mqtt_payload.configs)} configs')
            for config in ha_mqtt_payload.configs:
                self.publish(
                    topic=config['topic'],
                    payload=config['data'],
                )

        logger.debug(f'{log_prefix} send {len(ha_mqtt_payload.state["data"])} values')
        self.publish(
            topic=ha_mqtt_payload.state['topic'],
            payload=ha_mqtt_payload.state['data'],
        )
        self.send_count += 1



"""
    CLI for development
"""
import logging
import sys
from pathlib import Path

import rich_click as click
from bx_py_utils.path import assert_is_file
from manageprojects.utilities import code_style
from manageprojects.utilities.publish import publish_package
from manageprojects.utilities.subprocess_utils import verbose_check_call
from manageprojects.utilities.version_info import print_version
from rich import print  # noqa; noqa
from rich_click import RichGroup

import ha_services
from ha_services import constants


logger = logging.getLogger(__name__)


PACKAGE_ROOT = Path(ha_services.__file__).parent.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')

OPTION_ARGS_DEFAULT_TRUE = dict(is_flag=True, show_default=True, default=True)
OPTION_ARGS_DEFAULT_FALSE = dict(is_flag=True, show_default=True, default=False)
ARGUMENT_EXISTING_DIR = dict(
    type=click.Path(exists=True, file_okay=False, dir_okay=True, readable=True, path_type=Path)
)
ARGUMENT_NOT_EXISTING_DIR = dict(
    type=click.Path(
        exists=False,
        file_okay=False,
        dir_okay=True,
        readable=False,
        writable=True,
        path_type=Path,
    )
)
ARGUMENT_EXISTING_FILE = dict(
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, path_type=Path)
)


class ClickGroup(RichGroup):  # FIXME: How to set the "info_name" easier?
    def make_context(self, info_name, *args, **kwargs):
        info_name = './dev-cli.py'
        return super().make_context(info_name, *args, **kwargs)


@click.group(
    cls=ClickGroup,
    epilog=constants.CLI_EPILOG,
)
def cli():
    pass


@click.command()
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def mypy(verbose: bool = True):
    """Run Mypy (configured in pyproject.toml)"""
    verbose_check_call('mypy', '.', cwd=PACKAGE_ROOT, verbose=verbose, exit_on_error=True)


cli.add_command(mypy)


@click.command()
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def coverage(verbose: bool = True):
    """
    Run and show coverage.
    """
    verbose_check_call('coverage', 'run', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'combine', '--append', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'report', '--fail-under=30', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'xml', verbose=verbose, exit_on_error=True)
    verbose_check_call('coverage', 'json', verbose=verbose, exit_on_error=True)


cli.add_command(coverage)


@click.command()
def install():
    """
    Run pip-sync and install 'ha_services' via pip as editable.
    """
    verbose_check_call('pip-sync', PACKAGE_ROOT / 'requirements.dev.txt')
    verbose_check_call('pip', 'install', '--no-deps', '-e', '.')


cli.add_command(install)


@click.command()
def safety():
    """
    Run safety check against current requirements files
    """
    verbose_check_call('safety', 'check', '-r', 'requirements.dev.txt')


cli.add_command(safety)


@click.command()
def update():
    """
    Update "requirements*.txt" dependencies files
    """
    bin_path = Path(sys.executable).parent

    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip')
    verbose_check_call(bin_path / 'pip', 'install', '-U', 'pip-tools')

    extra_env = dict(
        CUSTOM_COMPILE_COMMAND='./cli.py update',
    )

    pip_compile_base = [
        bin_path / 'pip-compile',
        '--verbose',
        '--allow-unsafe',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--resolver=backtracking',  # https://pip-tools.readthedocs.io/en/latest/#deprecations
        '--upgrade',
        '--generate-hashes',
    ]

    # Only "prod" dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--output-file',
        'requirements.txt',
        extra_env=extra_env,
    )

    # dependencies + "dev"-optional-dependencies:
    verbose_check_call(
        *pip_compile_base,
        'pyproject.toml',
        '--extra=dev',
        '--output-file',
        'requirements.dev.txt',
        extra_env=extra_env,
    )

    verbose_check_call(bin_path / 'safety', 'check', '-r', 'requirements.dev.txt')

    # Install new dependencies in current .venv:
    verbose_check_call(bin_path / 'pip-sync', 'requirements.dev.txt')


cli.add_command(update)


@click.command()
def publish():
    """
    Build and upload this project to PyPi
    """
    _run_unittest_cli(verbose=False, exit_after_run=False)  # Don't publish a broken state

    publish_package(
        module=ha_services,
        package_path=PACKAGE_ROOT,
        distribution_name='ha-services',
    )


cli.add_command(publish)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def fix_code_style(color: bool = True, verbose: bool = False):
    """
    Fix code style of all ha_services source code files via darker
    """
    code_style.fix(package_root=PACKAGE_ROOT, color=color, verbose=verbose)


cli.add_command(fix_code_style)


@click.command()
@click.option('--color/--no-color', **OPTION_ARGS_DEFAULT_TRUE)
@click.option('--verbose/--no-verbose', **OPTION_ARGS_DEFAULT_FALSE)
def check_code_style(color: bool = True, verbose: bool = False):
    """
    Check code style by calling darker + flake8
    """
    code_style.check(package_root=PACKAGE_ROOT, color=color, verbose=verbose)


cli.add_command(check_code_style)


@click.command()
def update_test_snapshot_files():
    """
    Update all test snapshot files (by remove and recreate all snapshot files)
    """

    def iter_snapshot_files():
        yield from PACKAGE_ROOT.rglob('*.snapshot.*')

    removed_file_count = 0
    for item in iter_snapshot_files():
        item.unlink()
        removed_file_count += 1
    print(f'{removed_file_count} test snapshot files removed... run tests...')

    # Just recreate them by running tests:
    _run_unittest_cli(
        extra_env=dict(
            RAISE_SNAPSHOT_ERRORS='0',  # Recreate snapshot files without error
        ),
        verbose=False,
        exit_after_run=False,
    )

    new_files = len(list(iter_snapshot_files()))
    print(f'{new_files} test snapshot files created, ok.\n')


cli.add_command(update_test_snapshot_files)


def _run_unittest_cli(extra_env=None, verbose=True, exit_after_run=True):
    """
    Call the origin unittest CLI and pass all args to it.
    """
    if extra_env is None:
        extra_env = dict()

    extra_env.update(
        dict(
            PYTHONUNBUFFERED='1',
            PYTHONWARNINGS='always',
        )
    )

    args = sys.argv[2:]
    if not args:
        if verbose:
            args = ('--verbose', '--locals', '--buffer')
        else:
            args = ('--locals', '--buffer')

    verbose_check_call(
        sys.executable,
        '-m',
        'unittest',
        *args,
        timeout=15 * 60,
        extra_env=extra_env,
    )
    if exit_after_run:
        sys.exit(0)


@click.command()  # Dummy command
def test():
    """
    Run unittests
    """
    _run_unittest_cli()


cli.add_command(test)


def _run_tox():
    verbose_check_call(sys.executable, '-m', 'tox', *sys.argv[2:])
    sys.exit(0)


@click.command()  # Dummy "tox" command
def tox():
    """
    Run tox
    """
    _run_tox()


cli.add_command(tox)


@click.command()
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


cli.add_command(version)


def main():
    print_version(ha_services)

    if len(sys.argv) >= 2:
        # Check if we just pass a command call
        command = sys.argv[1]
        if command == 'test':
            _run_unittest_cli()
        elif command == 'tox':
            _run_tox()

    # Execute Click CLI:
    cli()



import tempfile
from pathlib import Path
from unittest import TestCase

from ha_services import __version__
from ha_services.systemd.data_classes import SystemdServiceInfo
from ha_services.systemd.tests.utilities import MockedSystemdServiceInfo


class SystemdDataClassesTestCase(TestCase):
    def test_systemd_service_info(self):
        info = SystemdServiceInfo()

        # Check some samples:
        self.assertEqual(info.template_context.verbose_service_name, 'HaServices Demo')
        self.assertEqual(info.template_context.version, __version__)
        self.assertEqual(info.service_slug, 'haservices_demo')
        self.assertEqual(info.template_context.syslog_identifier, 'haservices_demo')
        self.assertEqual(info.service_file_path, Path('/etc/systemd/system/haservices_demo.service'))

        with MockedSystemdServiceInfo(prefix='test_systemd_service_info_') as info:
            self.assertIsInstance(info, SystemdServiceInfo)
            self.assertEqual(info.template_context.user, 'MockedUserName')
            self.assertEqual(info.template_context.group, 'MockedUserName')
            self.assertEqual(
                info.template_context.exec_start, '/mocked/.venv/bin/python3 -m ha_services_app publish-loop'
            )

            base_temp_dir = tempfile.gettempdir()
            assert str(info.template_context.work_dir).startswith(
                base_temp_dir
            ), f'{info.template_context.work_dir} does not start with {base_temp_dir}'



import logging
import shutil
import subprocess


logger = logging.getLogger(__name__)


COMMANDS = (
    'sensible-editor',
    'mcedit',
    'nano',
    'edit',
    'open',
)


def open_editor_for(file_path):
    """
    Try to open the given file in a editor.
    """
    for command in COMMANDS:
        if bin := shutil.which(command):
            logger.debug('Call: "%s %s"', bin, file_path)
            try:
                return subprocess.check_call([bin, file_path])
            except subprocess.SubprocessError as err:
                print(f'Error open {file_path} with {bin}: [red]{err}')
        else:
            logger.debug(f'No "{command}" found.')

    print(f'Error not editor found to open {file_path}!')



import dataclasses
from pathlib import Path


@dataclasses.dataclass
class SimpleExample:
    """
    A simple example
    """

    one: str = 'foo'
    two: str = 'bar'
    three: str = ''
    number: int = 123


###########################################################################################


@dataclasses.dataclass
class PathExample:
    path: Path = Path('/foo/bar')


###########################################################################################


@dataclasses.dataclass
class SubClass1:
    """
    This is SubClass ONE on second level of FirstLevel
    """

    number: int = 123


@dataclasses.dataclass
class SubClass2:
    """
    This is SubClass TWO on second level of FirstLevel
    """

    something: float = 0.5


@dataclasses.dataclass
class SubClass3:
    one_value: bool = True


@dataclasses.dataclass
class ComplexExample:
    """
    This is the doc string, of the first level.
    It's two lines of doc string ;)
    """

    foo: str = 'bar'
    sub_class_one: dataclasses = dataclasses.field(default_factory=SubClass1)
    sub_class_two: dataclasses = dataclasses.field(default_factory=SubClass2)
    sub_class_three: dataclasses = dataclasses.field(default_factory=SubClass3)



import getpass
import logging
import os
import pwd
import shutil
from pathlib import Path

from bx_py_utils.environ import OverrideEnviron
from bx_py_utils.path import assert_is_file


logger = logging.getLogger(__name__)


def is_path_name(name: str) -> None:
    """
    >>> is_path_name('a_example_filename')
    >>> is_path_name('A Directory name')

    >>> is_path_name('/not/valid/')
    Traceback (most recent call last):
        ...
    AssertionError: Path name '/not/valid/' is not valid! (Not the same as: cleaned='valid')

    >>> is_path_name('invalid.exe')
    Traceback (most recent call last):
        ...
    AssertionError: Path name 'invalid.exe' is not valid! (Not the same as: cleaned='invalid')
    """
    cleaned = Path(name).stem
    if name != cleaned:
        raise AssertionError(f'Path name {name!r} is not valid! (Not the same as: {cleaned=!r})')


def expand_user(path: Path) -> Path:
    """
    Returns a new path with expanded ~ and ~user constructs:
    Unlike the normal Python function, when called with sudo, the normal user path is used.
    So "~" is not expanded as "/root" -> it's expanded with the user home that sudo starts with!
    """
    logger.debug(f'expand user path: {path}')
    if sudo_user := os.environ.get('SUDO_USER'):
        if sudo_user == 'root':
            logger.warning('Do not run this as root user! Please use a normal user and sudo!')

        env_user = getpass.getuser()
        logger.debug(f'SUDO_USER:{sudo_user!r} <-> {env_user}')
        if sudo_user != env_user:
            # Get home directory of the user that starts sudo via password database:
            sudo_user_home = pwd.getpwnam(sudo_user).pw_dir
            with OverrideEnviron(HOME=sudo_user_home):
                return Path(path).expanduser()
    else:
        return Path(path).expanduser()


def backup(file_path: Path, max_try=100) -> Path:
    """
    Backup the given file, by create copy with the suffix: ".bak"
    Increment a number to the suffix -> So no old backup file will be overwritten.
    """
    assert_is_file(file_path)
    for number in range(1, max_try + 1):
        bak_file_candidate = file_path.with_name(f'{file_path.name}.bak{number if number>1 else ""}')
        if not bak_file_candidate.is_file():
            logger.info('Backup %s to %s', file_path, bak_file_candidate)
            shutil.copyfile(file_path, bak_file_candidate)
            return bak_file_candidate
    raise RuntimeError('No backup made: Maximum attempts to find a file name failed.')



import os
import unittest.util

from ha_services.log_setup import basic_log_setup


# Hacky way to expand the failed test output:
unittest.util._MAX_LENGTH = os.environ.get('UNITTEST_MAX_LENGTH', 300)


# Display DEBUG logs in tests:
basic_log_setup(debug=True)



from ha_services.mqtt4homeassistant.data_classes import HaMqttPayload, HaValue, HaValues
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify


def values2mqtt_payload(
    values: HaValues,
    name_prefix: str,
    default_device_class: str = 'energy',
    default_state_class: str = 'measurement',
) -> HaMqttPayload:
    assert not name_prefix[0].isdigit(), f'Invalid: {name_prefix=}'
    device_id = f'{slugify(name_prefix).lower()}_{slugify(values.device_name).lower()}'
    state = {
        'data': {},
        'topic': f'{values.prefix}/{values.component}/{device_id}/state',
    }
    config_data = {}
    for value in values.values:
        assert isinstance(value, HaValue), f'Wrong type: {value=!r}'

        key = slugify(value.name).lower()
        unique_id = f'{device_id}_{key}'

        assert key not in config_data, f'Double {unique_id=!r} from {value=!r} - {config_data=!r}'
        config_data[unique_id] = dict(
            name=value.name,
            device_class=value.device_class or default_device_class,
            state_class=value.state_class or default_state_class,
            state_topic=state['topic'],
            unit_of_measurement=value.unit,
            unique_id=unique_id,
            value_template='{{ value_json.%(value_key)s }}' % dict(value_key=unique_id),
        )
        state['data'][unique_id] = value.value

    configs = []
    for unique_id, data in config_data.items():
        configs.append(
            {
                'data': {
                    'device': {
                        'identifiers': sorted(config_data.keys()),
                        'name': values.device_name,
                    },
                    **data,
                },
                'topic': f'{values.prefix}/{values.component}/{unique_id}/config',
            }
        )

    return HaMqttPayload(configs=configs, state=state)



import getpass
import os
import sys
from pathlib import Path

from ha_services.constants import BASE_PATH


def get_template_path() -> Path:
    return BASE_PATH / 'systemd' / 'service_template.txt'


def get_user_name() -> str:
    user_name = os.environ.get('SUDO_USER', getpass.getuser())
    return user_name


get_user_group = get_user_name


def get_work_directory() -> Path:
    return Path().cwd()


def get_demo_exec_start() -> str:
    return f'{sys.executable} -m ha_services_app publish-loop'



import os
import tempfile
from pathlib import Path
from unittest import mock

from bx_py_utils.environ import OverrideEnviron
from bx_py_utils.test_utils.context_managers import MassContextManager

from ha_services.systemd import defaults
from ha_services.systemd.data_classes import SystemdServiceInfo


class MockedCurrentWorkDir(tempfile.TemporaryDirectory):
    def __init__(self, **kwargs):
        self.old_cwd = Path().cwd()
        super().__init__(**kwargs)

    def __enter__(self):
        temp_dir = super().__enter__()
        os.chdir(temp_dir)
        self.temp_path = Path(temp_dir)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        os.chdir(self.old_cwd)


class MockedSys:
    executable = '/mocked/.venv/bin/python3'


class MockedSystemdServiceInfo(MassContextManager):
    """
    Set all values in ha_services/systemd/defaults.py to static ones,
    independent of current user/environment.

    So that creating a SystemdServiceInfo() instance will result in a well known state.

    Note: The work_dir has still a random suffix!
    """

    def __init__(self, prefix: str):
        self.mocked_cwd = MockedCurrentWorkDir(prefix=prefix)
        self.mocks = (
            self.mocked_cwd,
            OverrideEnviron(SUDO_USER='MockedUserName'),
            mock.patch.object(defaults, 'sys', MockedSys()),
        )

    def __enter__(self) -> SystemdServiceInfo:
        super().__enter__()
        temp_path = self.mocked_cwd.temp_path
        mocked_systemd_base_path = temp_path / 'etc-systemd-system'
        mocked_systemd_base_path.mkdir()
        info = SystemdServiceInfo(systemd_base_path=mocked_systemd_base_path)
        return info



from unittest import TestCase

from ha_services.systemd.template import InvalidTemplate, TemplateKeyInfo, get_template_key_info, validate_template


class TemplateTestCase(TestCase):
    def test_validate_template(self):
        content = 'One $foo and ${bar} ?!?'
        context = dict(foo='FOO', bar='BAR')
        self.assertEqual(
            get_template_key_info(content=content, context=context),
            TemplateKeyInfo(template_keys=['bar', 'foo'], not_in_context=[], not_in_template=[], invalid_keys=[]),
        )
        validate_template(content=content, context=context)

        ##########################################################################################

        context = dict(foo='FOO')
        self.assertEqual(
            get_template_key_info(content=content, context=context),
            TemplateKeyInfo(template_keys=['bar', 'foo'], not_in_context=['bar'], not_in_template=[], invalid_keys=[]),
        )
        with self.assertRaises(InvalidTemplate) as cm:
            validate_template(content=content, context=context)
        self.assertEqual(str(cm.exception), 'Template key(s) "bar" not in context')

        ##########################################################################################

        content = 'One $foo no bar!'
        context = dict(foo='FOO', bar='BAR')
        self.assertEqual(
            get_template_key_info(content=content, context=context),
            TemplateKeyInfo(template_keys=['foo'], not_in_context=[], not_in_template=['bar'], invalid_keys=[]),
        )
        with self.assertRaises(InvalidTemplate) as cm:
            validate_template(content=content, context=context)
        self.assertEqual(str(cm.exception), 'Context key(s) "bar" not in template')

        ##########################################################################################

        content = 'One $foo and ${bar} ?!?'
        context = dict(foo='FOO', baz='666')
        self.assertEqual(
            get_template_key_info(content=content, context=context),
            TemplateKeyInfo(
                template_keys=['bar', 'foo'], not_in_context=['bar'], not_in_template=['baz'], invalid_keys=[]
            ),
        )
        with self.assertRaises(InvalidTemplate) as cm:
            validate_template(content=content, context=context)
        self.assertEqual(
            str(cm.exception), 'Template key(s) "bar" not in context & Context key(s) "baz" not in template'
        )



import dataclasses


def iter_dataclass(instance):
    for field in dataclasses.fields(instance):
        field_name = field.name
        field_value = getattr(instance, field_name)
        yield field_name, field_value



import dataclasses
import logging
import os
import resource
import time

from rich import print  # noqa

from ha_services.mqtt4homeassistant.converter import values2mqtt_payload
from ha_services.mqtt4homeassistant.data_classes import HaValue, HaValues, MqttSettings
from ha_services.mqtt4homeassistant.mqtt import HaMqttPublisher
from ha_services.systemd.data_classes import SystemdServiceInfo


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class MqttExampleValues:
    """
    Some values used to create DEMO MQTT messages.
    """

    mqtt_payload_prefix: str = 'example'
    device_name: str = 'ha-services-demo'


@dataclasses.dataclass
class DemoSettings:
    """
    This are just settings for the "ha-services" DEMO.
    Will be used in ha_services example commands.
    See "./cli.py --help" for more information.
    """

    mqtt: dataclasses = dataclasses.field(default_factory=MqttSettings)
    app: dataclasses = dataclasses.field(default_factory=MqttExampleValues)
    systemd: dataclasses = dataclasses.field(default_factory=SystemdServiceInfo)


def publish_forever(*, user_settings: DemoSettings, verbose):
    publisher = HaMqttPublisher(
        settings=user_settings.mqtt,
        verbose=verbose,
        config_count=1,  # Send every time the config
    )

    while True:
        # Just collect something that we can send:
        usage = resource.getrusage(resource.RUSAGE_SELF)
        values = [
            HaValue(
                name='System load 1min.',
                value=os.getloadavg()[0],
                device_class='',
                state_class='measurement',
                unit='',
            ),
            HaValue(
                name='Time in user mode (float seconds)',
                value=usage.ru_utime,
                device_class='',
                state_class='measurement',
                unit='sec',
            ),
            HaValue(
                name='Time in system mode (float seconds)',
                value=usage.ru_stime,
                device_class='',
                state_class='measurement',
                unit='sec',
            ),
        ]

        # Collect information:
        ha_values = HaValues(
            device_name=user_settings.app.device_name,
            values=values,
        )

        # Create Payload:
        ha_mqtt_payload = values2mqtt_payload(values=ha_values, name_prefix=user_settings.app.mqtt_payload_prefix)

        # Send vial MQTT to HomeAssistant:
        publisher.publish2homeassistant(ha_mqtt_payload=ha_mqtt_payload)

        print('Wait', end='...')
        for i in range(10, 1, -1):
            time.sleep(1)
            print(i, end='...')



import dataclasses
import inspect
from unittest import TestCase

import tomlkit
from tomlkit import TOMLDocument

from ha_services.toml_settings.serialize import dataclass2toml
from ha_services.toml_settings.tests.fixtures import ComplexExample, PathExample, SimpleExample


class SerializeTestCase(TestCase):
    def test_dataclass2toml_simple(self):
        document = dataclass2toml(instance=SimpleExample())
        self.assertIsInstance(document, TOMLDocument)
        self.assertEqual(document.unwrap(), {'one': 'foo', 'two': 'bar', 'three': '', 'number': 123})

        doc_str = tomlkit.dumps(document, sort_keys=False).rstrip()
        self.assertEqual(
            doc_str,
            inspect.cleandoc(
                '''
                # A simple example
                one = "foo"
                two = "bar"
                three = ""
                number = 123
                '''
            ),
        )

    def test_dataclass2toml_path(self):
        document = dataclass2toml(instance=PathExample())
        self.assertIsInstance(document, TOMLDocument)
        self.assertEqual(document.unwrap(), {'path': '/foo/bar'})

        doc_str = tomlkit.dumps(document, sort_keys=False).rstrip()
        self.assertEqual(
            doc_str,
            inspect.cleandoc(
                '''
                # PathExample(path: pathlib.Path = PosixPath('/foo/bar'))
                path = "/foo/bar"
                '''
            ),
        )

    def test_dataclass2toml_inheritance(self):
        instance = ComplexExample()
        data = dataclasses.asdict(instance)
        self.assertEqual(
            data,
            {
                'foo': 'bar',
                'sub_class_one': {'number': 123},
                'sub_class_two': {'something': 0.5},
                'sub_class_three': {'one_value': True},
            },
        )

        document = dataclass2toml(instance=instance)
        self.assertIsInstance(document, TOMLDocument)
        self.assertEqual(document.unwrap(), data)

        doc_str = tomlkit.dumps(document, sort_keys=False).rstrip()
        self.assertEqual(
            doc_str,
            inspect.cleandoc(
                '''
                # This is the doc string, of the first level.
                # It's two lines of doc string ;)
                foo = "bar"

                [sub_class_one]
                # This is SubClass ONE on second level of FirstLevel
                number = 123

                [sub_class_two]
                # This is SubClass TWO on second level of FirstLevel
                something = 0.5

                [sub_class_three]
                # SubClass3(one_value: bool = True)
                one_value = true
                '''
            ),
        )



import dataclasses
import inspect
from pathlib import Path

import tomlkit
from tomlkit import TOMLDocument
from tomlkit.items import Item

from ha_services.toml_settings.data_class_utils import iter_dataclass


def add_docstring(document: TOMLDocument, instance):
    if doc_string := inspect.getdoc(instance):
        for line in doc_string.strip().splitlines():
            document.add(tomlkit.comment(line))


def add_value(*, field_name, item: Item, value):
    if isinstance(value, Path):
        value = str(value)

    item.add(field_name, value)


def add_dataclass(document: TOMLDocument, name, instance):
    assert dataclasses.is_dataclass(instance), f'No dataclass: {instance!r}'

    document.add(tomlkit.nl())  # Add new line

    table = tomlkit.table()
    add_docstring(table, instance)

    for field_name, field_value in iter_dataclass(instance):
        if dataclasses.is_dataclass(field_value):
            add_dataclass(table, field_name, field_value)
        else:
            add_value(field_name=field_name, item=table, value=field_value)

    document.add(name, table)


def dataclass2toml(instance) -> TOMLDocument:
    """
    Serialize a dataclass to a tomlkit document.
    """
    assert dataclasses.is_dataclass(instance), f'No dataclass: {instance!r}'

    document = tomlkit.document()
    add_docstring(document, instance)

    for field_name, field_value in iter_dataclass(instance):
        if dataclasses.is_dataclass(field_value):
            add_dataclass(document, field_name, field_value)
        else:
            add_value(field_name=field_name, item=document, value=field_value)

    return document



import getpass
import os
from pathlib import Path
from unittest import TestCase

from ha_services.cli_tools.test_utils.environment_fixtures import AsSudoCallOverrideEnviron


class EnvironFixturesTestCase(TestCase):
    def test_happy_path(self):
        user_id = os.getuid()
        user_group_id = os.getgid()

        with AsSudoCallOverrideEnviron():
            self.assertEqual(getpass.getuser(), 'root')
            self.assertEqual(Path().home(), Path('/root'))
            self.assertEqual(Path('~/foo/bar/').expanduser(), Path('/root/foo/bar'))

            self.assertNotEqual(os.environ['SUDO_UID'], user_id)
            self.assertNotEqual(os.environ['SUDO_GID'], user_group_id)



from __future__ import annotations

import dataclasses

from bx_py_utils.anonymize import anonymize


@dataclasses.dataclass
class HaValue:
    name: str
    value: int | float | str
    device_class: str  # e.g.: "voltage" / "current" / "energy" etc.
    state_class: str  # e.g.: "measurement" / "total" / "total_increasing" etc.
    unit: str | None  # e.g.: "V" / "A" / "kWh" etc.


@dataclasses.dataclass
class HaValues:
    device_name: str
    values: list[HaValue, ...]
    prefix: str = 'homeassistant'
    component: str = 'sensor'


@dataclasses.dataclass
class HaMqttPayload:
    configs: list[dict, ...]
    state: dict


@dataclasses.dataclass
class MqttSettings:
    """
    Credentials to MQTT server that should be used.
    """

    host: str = 'mqtt.eclipseprojects.io'  # public test MQTT broker service
    port: int = 1883
    user_name: str = ''
    password: str = ''

    def anonymized(self):
        data = dataclasses.asdict(self)
        if self.password:
            data['password'] = anonymize(self.password)
        return data
